/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of TinyPLC
 *
 *  TinyPLC is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; version 2 of the License.
 *
 *  TinyPLC is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/*
 *  My thanks to my employer, Athena Controls (www.athenacontrols.com)
 *  for allowing me to open source this project.
 */
/*
 *  TinyPLC kernel
 *  Processor:   Atmel MEGA1281 or Atmel MEGA2561
 *  Compiler:    Codevision AVR C compiler, 1.24.8e
 *
 *  TinyPLC editor/compiler
 *  Processor:   PC-clone, VESA video (mode 0x105, 1024x768x256)
 *  Compiler:    DeSmet C, version 2.51 (PCC version 1.2)
 *
 *  Revision history:
 *  Programmer        Date       Comments
 *  ----------------  ---------  ---------------------------------------------
 *  William Couture   3/19/2006  Original code
 *  William Couture   9/13/2006  Modify for GPL release
 */

#include "plc.h"

int getnumber(str,val)
   BYTE *str;
   DWORD *val;
   {
   WORD ishex;
   WORD isneg;
   DWORD v;
   DWORD digit;

   isneg = FALSE;
   ishex = FALSE;
   if ((str[0] == '0') && (toupper(str[1]) == 'X'))
      {
      ishex = TRUE;
      str += 2;
      }
   v = 0;
   if (str[0] == '-')
      {
      isneg = TRUE;
      str++;
      }
   if (!str[0])
      return(FALSE);  /* empty string */
   if ((isneg) && (ishex))
      return(FALSE);  /* negative hex won't fly here... */
   while (*str)
      {
      *str = toupper(*str);
      if ((*str >= '0') && (*str <= '9'))
         digit = *str - '0';
      else
         if ((ishex) && (*str >= 'A') && (*str <= 'F'))
            digit = *str - 'A' + 10;
         else
            return(FALSE);
      if (ishex)
         v <<= 4;
      else
         v *= 10;
      v += digit;
      str++;
      }
   if (isneg)
      v = -v;
   *val = v;
   return(TRUE);
   }

int parse_io_str(str,type,reg)
   BYTE *str, *type;
   DWORD *reg;
   {
   BYTE ch;
   DWORD val;

   *type = UNUSED_TYPE;  /* default to FALSE */
   *reg = 0;
   if (!str[0])
      return(FALSE);
   if ((ch = toupper(str[0])) == INPUT_PREFIX)
      {
      if ((getnumber(&str[1],&val)) && (!(val & 0x80000000L))
                     && (val < NUM_INPUTS))
         {
         *type = INPUT_TYPE;
         *reg = FIRST_INPUT + val;
         return(TRUE);
         }
      return(beep());
      }
   if ((ch = toupper(str[0])) == VAR_PREFIX)
      {
      if ((getnumber(&str[1],&val)) && (!(val & 0x80000000L))
                     && (val < NUM_VARS))
         {
         *type = VAR_TYPE;
         *reg = FIRST_VAR + val;
         return(TRUE);
         }
      return(beep());
      }
   if ((ch = toupper(str[0])) == TIMER_PREFIX)
      {
      if ((getnumber(&str[1],&val)) && (!(val & 0x80000000L))
                     && (val < NUM_TIMERS))
         {
         *type = TIMER_TYPE;
         *reg = FIRST_TIMER + (val << 1);
         return(TRUE);
         }
      return(beep());
      }
   if ((ch = toupper(str[0])) == COUNTER_PREFIX)
      {
      if ((getnumber(&str[1],&val)) && (!(val & 0x80000000L))
                     && (val < NUM_COUNTERS))
         {
         *type = COUNTER_TYPE;
         *reg = FIRST_COUNTER + val;
         return(TRUE);
         }
      return(beep());
      }
   if ((ch = toupper(str[0])) == ONESHOT_PREFIX)
      {
      if ((getnumber(&str[1],&val)) && (!(val & 0x80000000L))
                     && (val < NUM_ONESHOTS))
         {
         *type = ONESHOT_TYPE;
         *reg = FIRST_ONESHOT + val;
         return(TRUE);
         }
      return(beep());
      }
   if ((ch = toupper(str[0])) == ANALOG_PREFIX)
      {
      if ((getnumber(&str[1],&val)) && (!(val & 0x80000000L))
                     && (val < NUM_ANALOGS))
         {
         *type = ANALOG_TYPE;
         *reg = FIRST_ANALOG + val;
         return(TRUE);
         }
      return(beep());
      }
   if ((ch = toupper(str[0])) == OUTPUT_PREFIX)
      {
      if ((getnumber(&str[1],&val)) && (!(val & 0x80000000L))
                     && (val < NUM_OUTPUTS))
         {
         *type = OUTPUT_TYPE;
         *reg = FIRST_OUTPUT + val;
         return(TRUE);
         }
      return(beep());
      }
   if ((ch = toupper(str[0])) == LABEL_PREFIX)
      {
      if ((getnumber(&str[1],&val)) && (!(val & ~255L)))
         {
         *type = LABEL_TYPE;
         *reg = (BYTE)val;
         return(TRUE);
         }
      return(beep());
      }
   if (getnumber(str,&val))
      {
      if (val & ~255L)
         {
         if (val & ~65535L)
            {
            *type = DWORDVAL_TYPE;
            *reg = val;
            return(TRUE);
            }
         *type = WORDVAL_TYPE;
         *reg = (WORD)val;
         return(TRUE);
         }
      *type = BYTEVAL_TYPE;
      *reg = (BYTE)val;
      return(TRUE);
      }
   return(beep());
   }

BYTE *logicnodetype_names[NUM_VAR_TYPES] =
   {
   NULL,
   "X",
   "V",
   "T",
   "C",
   "R",
   "A",
   "Y",
   "L",
   NULL,
   NULL,
   NULL
   };

WORD regbase[NUM_VAR_TYPES] =
   {
   0,
   FIRST_INPUT,
   FIRST_VAR,
   FIRST_TIMER,
   FIRST_COUNTER,
   FIRST_ONESHOT,
   FIRST_ANALOG,
   FIRST_OUTPUT,
   0,
   0,
   0,
   0
   };

showregtype(r,c,type,reg)
   WORD r, c;
   BYTE type;
   DWORD reg;
   {
   BYTE astr[16];

   if (type == UNUSED_TYPE)
      {
      smallstrout(r,c,"     ");  /* make sure previous content is erased */
      smallstrout(r,c,"?");
      }
   else
      {
      if (type == BYTEVAL_TYPE)
         sprintf(astr,"%d",reg & 0xff);
      else
         if (type == WORDVAL_TYPE)
            sprintf(astr,"%u",reg & 0xffff);
         else
            if (type == DWORDVAL_TYPE)
               sprintf(astr,"%ld",reg);
            else
               if (type == TIMER_TYPE)
                  sprintf(astr,"%s%02d",logicnodetype_names[type],((reg & 0xff)-regbase[type]) >> 1);
               else
                  sprintf(astr,"%s%02d",logicnodetype_names[type],(reg & 0xff)-regbase[type]);
      smallstrout(r,c,astr);
      }
   }

void erase_node_notes(r,c)
   {
   smallstrout(r-TEXTHEIGHTSMALL,c,"        ");
   smallstrout(r+TEXTHEIGHT,c,"        ");
   }

void clear_node(nodeptr)
   struct logic_element *nodeptr;
   {
   WORD i;

   for (i = 0; i < OPCODE_MAX; i++)
      {
      nodeptr->opcode[i] = 0;
      nodeptr->opflags[i] = UNUSED_TYPE;
      }
   }

WORD compile_node(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD len;
   BYTE *buf;
   {
   WORD i;

   for (i = 0; i < len; i++)
      {
      buf[i] = nodeptr->opcode[i];
      if ((i) && (nodeptr->opflags[i] == UNUSED_TYPE))  /* input type not known */
         return(COMPILE_UNKNOWN_PARAMS);
      }
   return(COMPILE_OK);
   }

WORD compile_endian_node(nodeptr,ofs,buf,len)
   struct logic_element *nodeptr;
   WORD ofs;
   BYTE *buf;
   WORD len;
   {
   WORD i;

   for (i = 0; i < ofs; i++)
      {
      buf[i] = nodeptr->opcode[i];
      if ((i) && (nodeptr->opflags[i] == UNUSED_TYPE))  /* input type not known */
         return(COMPILE_UNKNOWN_PARAMS);
      }
   if ((nodeptr->opflags[ofs] != WORDVAL_TYPE) && (nodeptr->opflags[ofs] != DWORDVAL_TYPE))
      return(COMPILE_UNKNOWN_PARAMS);
   if (is_littleendian)
      {
      buf[ofs] = nodeptr->opcode[ofs];  /* little-endian */
      buf[ofs+1] = nodeptr->opcode[ofs+1];
      if (nodeptr->opflags[ofs] == DWORDVAL_TYPE)
         {
         buf[ofs+2] = nodeptr->opcode[ofs+2];
         buf[ofs+3] = nodeptr->opcode[ofs+3];
         }
      }
   else
      {
      if (nodeptr->opflags[ofs] == DWORDVAL_TYPE)
         {
         buf[ofs] = nodeptr->opcode[ofs+3];  /* big-endian */
         buf[ofs+1] = nodeptr->opcode[ofs+2];
         buf[ofs+2] = nodeptr->opcode[ofs+1];
         buf[ofs+3] = nodeptr->opcode[ofs];
         }
      else
         {
         buf[ofs] = nodeptr->opcode[ofs+1];  /* big-endian */
         buf[ofs+1] = nodeptr->opcode[ofs];
         }
      }
   return(COMPILE_OK);
   }

void inp_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("N/O Input","Input ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if (type != LABEL_TYPE)
         {
         nodeptr->opcode[0] = INP;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void inp_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void inp_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[INP_OBJ].label,fg,bg);
   coverlay(r,c,objs[INP_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+2*TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD inp_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 2;
   return(compile_node(nodeptr,*len,buf));
   }

void ninp_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("N/C Input","Input ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if (type != LABEL_TYPE)
         {
         nodeptr->opcode[0] = NINP;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void ninp_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void ninp_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[NINP_OBJ].label,fg,bg);
   coverlay(r,c,objs[NINP_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+2*TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD ninp_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 2;
   return(compile_node(nodeptr,*len,buf));
   }

void out_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("N/O Output","Output ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if ((type == VAR_TYPE) || (type == OUTPUT_TYPE))
         {
         nodeptr->opcode[0] = OUT;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void out_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void out_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[OUT_OBJ].label,fg,bg);
   coverlay(r,c,objs[OUT_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+2*TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD out_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 2;
   return(compile_node(nodeptr,*len,buf));
   }

void nout_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("N/C Output","Output ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if ((type == VAR_TYPE) || (type == OUTPUT_TYPE))
         {
         nodeptr->opcode[0] = NOUT;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void nout_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void nout_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[NOUT_OBJ].label,fg,bg);
   coverlay(r,c,objs[NOUT_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+2*TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD nout_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 2;
   return(compile_node(nodeptr,*len,buf));
   }

void lout_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("Latch Output","Output ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if ((type == VAR_TYPE) || (type == OUTPUT_TYPE))
         {
         nodeptr->opcode[0] = LOUT;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void lout_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void lout_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[LOUT_OBJ].label,fg,bg);
   coverlay(r,c,objs[LOUT_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD lout_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 2;
   return(compile_node(nodeptr,*len,buf));
   }

void uout_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("Unlatch Output","Output ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if ((type == VAR_TYPE) || (type == OUTPUT_TYPE))
         {
         nodeptr->opcode[0] = UOUT;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void uout_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void uout_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[UOUT_OBJ].label,fg,bg);
   coverlay(r,c,objs[UOUT_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD uout_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 2;
   return(compile_node(nodeptr,*len,buf));
   }

void math_node(nodeptr, label, opcode, opcodeb,swapok)
   struct logic_element *nodeptr;
   BYTE *label, opcode, opcodeb;
   {
   BYTE str1[12], str2[12], strdest[12];
   BYTE type1,type2,typedest;
   DWORD reg1,reg2,regdest;
   DWORD tempreg;

   get_values(label,
              "Destination ? ",
              "First op    ? ",
              "Second op   ? ",
              strdest,
              str1,
              str2);
   if ((parse_io_str(str1,&type1,&reg1))
          && (parse_io_str(str2,&type2,&reg2))
          && (parse_io_str(strdest,&typedest,&regdest)))
      {
      if ((typedest != VAR_TYPE) && (typedest != OUTPUT_TYPE))
         {
         beep();
         return;
         }
      if ((type1 == LABEL_TYPE) || (type2 == LABEL_TYPE))
         {
         beep();
         return;
         }
      if (type1 == BYTEVAL_TYPE)
         {
         if (!swapok)
            {
            beep();
            return;
            }
         if (type2 == BYTEVAL_TYPE)
            {
            beep();
            return;
            }
         tempreg = reg1;
         type1 = type2;
         reg1 = reg2;
         type2 = BYTEVAL_TYPE;
         reg2 = tempreg;
         }
      nodeptr->opcode[0] = (type2 == BYTEVAL_TYPE) ? opcodeb : opcode;
      nodeptr->opcode[1] = reg1;
      nodeptr->opcode[2] = reg2;
      nodeptr->opcode[3] = regdest;
      nodeptr->opflags[1] = type1;
      nodeptr->opflags[2] = type2;
      nodeptr->opflags[3] = typedest;
      }
   }

math_draw(r,c,nodeptr,symbol)
   WORD r, c;
   struct logic_element *nodeptr;
   BYTE *symbol;
   {
   showregtype(r-TEXTHEIGHTSMALL,c,nodeptr->opflags[3],(DWORD)nodeptr->opcode[3]);
   smallstrout(r-TEXTHEIGHTSMALL,c+4*TEXTWIDTH,"=");
   showregtype(r+TEXTHEIGHT,c,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   smallstrout(r+TEXTHEIGHT,c+3*TEXTWIDTH,symbol);
   showregtype(r+TEXTHEIGHT,c+4*TEXTWIDTH,nodeptr->opflags[2],(DWORD)nodeptr->opcode[2]);
   }

void add_init(nodeptr)
   struct logic_element *nodeptr;
   {
   math_node(nodeptr,"Add: dest = (1) + (2)",ADD,ADDB,TRUE);
   }

void add_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void add_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[ADD_OBJ].label,fg,bg);
   coverlay(r,c,objs[ADD_OBJ].overlay,fg);
   math_draw(r,c,nodeptr,"+");
   }

WORD add_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_node(nodeptr,*len,buf));
   }

void sub_init(nodeptr)
   struct logic_element *nodeptr;
   {
   math_node(nodeptr,"Sub: dest = (1) - (2)",SUB,SUBB,FALSE);
   }

void sub_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void sub_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[SUB_OBJ].label,fg,bg);
   coverlay(r,c,objs[SUB_OBJ].overlay,fg);
   math_draw(r,c,nodeptr,"-");
   }

WORD sub_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_node(nodeptr,*len,buf));
   }

void mul_init(nodeptr)
   struct logic_element *nodeptr;
   {
   math_node(nodeptr,"Mul: dest = (1) * (2)",MUL,MULB,TRUE);
   }

void mul_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void mul_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[MUL_OBJ].label,fg,bg);
   coverlay(r,c,objs[MUL_OBJ].overlay,fg);
   math_draw(r,c,nodeptr,"*");
   }

WORD mul_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_node(nodeptr,*len,buf));
   }

void div_init(nodeptr)
   struct logic_element *nodeptr;
   {
   WORD i;

   math_node(nodeptr,"Div: dest = (1) / (2)",DIV,DIVB,FALSE);
   if ((nodeptr->opflags[2] == BYTEVAL_TYPE) && (nodeptr->opcode[2] == 0))
      {
      for (i = 1; i < OPCODE_MAX; i++)
         nodeptr->opflags[i] = UNUSED_TYPE;
      beep();
      }
   }

void div_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void div_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[DIV_OBJ].label,fg,bg);
   coverlay(r,c,objs[DIV_OBJ].overlay,fg);
   math_draw(r,c,nodeptr,"/");
   }

WORD div_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_node(nodeptr,*len,buf));
   }

void mod_init(nodeptr)
   struct logic_element *nodeptr;
   {
   WORD i;

   math_node(nodeptr,"Mod: dest = (1) % (2)",MOD,MODB,FALSE);
   if ((nodeptr->opflags[2] == BYTEVAL_TYPE) && (nodeptr->opcode[2] == 0))
      {
      for (i = 1; i < OPCODE_MAX; i++)
         nodeptr->opflags[i] = UNUSED_TYPE;
      beep();
      }
   }

void mod_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void mod_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[MOD_OBJ].label,fg,bg);
   coverlay(r,c,objs[MOD_OBJ].overlay,fg);
   math_draw(r,c,nodeptr,"%");
   }

WORD mod_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_node(nodeptr,*len,buf));
   }

void and_init(nodeptr)
   struct logic_element *nodeptr;
   {
   math_node(nodeptr,"AND: dest = (1) & (2)",AND,ANDB,TRUE);
   }

void and_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void and_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[AND_OBJ].label,fg,bg);
   coverlay(r,c,objs[AND_OBJ].overlay,fg);
   math_draw(r,c,nodeptr,"&");
   }

WORD and_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_node(nodeptr,*len,buf));
   }

void or_init(nodeptr)
   struct logic_element *nodeptr;
   {
   math_node(nodeptr,"OR: dest = (1) | (2)",OR,ORB,TRUE);
   }

void or_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void or_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[OR_OBJ].label,fg,bg);
   coverlay(r,c,objs[OR_OBJ].overlay,fg);
   math_draw(r,c,nodeptr,"|");
   }

WORD or_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_node(nodeptr,*len,buf));
   }

void xor_init(nodeptr)
   struct logic_element *nodeptr;
   {
   math_node(nodeptr,"XOR: dest = (1) ^ (2)",XOR,XORB,TRUE);
   }

void xor_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void xor_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[XOR_OBJ].label,fg,bg);
   coverlay(r,c,objs[XOR_OBJ].overlay,fg);
   math_draw(r,c,nodeptr,"^");
   }

WORD xor_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_node(nodeptr,*len,buf));
   }

void not_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("NOT Output","Output / Var ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if ((type == VAR_TYPE) || (type == OUTPUT_TYPE))
         {
         nodeptr->opcode[0] = NOT;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void not_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void not_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[NOT_OBJ].label,fg,bg);
   coverlay(r,c,objs[NOT_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD not_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 2;
   return(compile_node(nodeptr,*len,buf));
   }

void neg_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("NEG variable","Variable ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if (type == VAR_TYPE)
         {
         nodeptr->opcode[0] = NEG;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void neg_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void neg_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[NEG_OBJ].label,fg,bg);
   coverlay(r,c,objs[NEG_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD neg_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 2;
   return(compile_node(nodeptr,*len,buf));
   }

void cmp_node(nodeptr, label, opcode, opcodeb, swapop, swapopb, swapobj)
   struct logic_element *nodeptr;
   BYTE *label, opcode, opcodeb, swapop, swapopb;
   WORD swapobj;
   {
   BYTE str1[12], str2[12];
   BYTE type1,type2;
   DWORD reg1,reg2;
   DWORD tempreg;

   get_values(label,
              "First op    ? ",
              "Second op   ? ",
              NULL,
              str1,
              str2,
              NULL);
   if ((parse_io_str(str1,&type1,&reg1))
          && (parse_io_str(str2,&type2,&reg2)))
      {
      if ((type1 == LABEL_TYPE) || (type2 == LABEL_TYPE))
         {
         beep();
         return;
         }
      if (type1 == BYTEVAL_TYPE)
         {
         if (type2 == BYTEVAL_TYPE)
            {
            return;
            beep();
            }
         tempreg = reg1;
         type1 = type2;
         reg1 = reg2;
         type2 = BYTEVAL_TYPE;
         reg2 = tempreg;
         opcode = swapop;
         opcodeb = swapopb;
         nodeptr->objindex = swapobj;
         }
      nodeptr->opcode[0] = (type2 == BYTEVAL_TYPE) ? opcodeb : opcode;
      nodeptr->opcode[1] = reg1;
      nodeptr->opcode[2] = reg2;
      nodeptr->opflags[1] = type1;
      nodeptr->opflags[2] = type2;
      }
   }

cmp_draw(r,c,nodeptr,symbol)
   WORD r, c;
   struct logic_element *nodeptr;
   BYTE *symbol;
   {
   showregtype(r-TEXTHEIGHTSMALL,c,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   smallstrout(r-TEXTHEIGHTSMALL,c+4*TEXTWIDTH,symbol);
   showregtype(r+TEXTHEIGHT,c+2*TEXTWIDTH,nodeptr->opflags[2],(DWORD)nodeptr->opcode[2]);
   }

void equ_init(nodeptr)
   struct logic_element *nodeptr;
   {
   cmp_node(nodeptr, "EQU:  (1) == (2)", EQU, EQUB, EQU, EQUB, EQU_OBJ);
   }

void equ_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void equ_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[EQU_OBJ].label,fg,bg);
   coverlay(r,c,objs[EQU_OBJ].overlay,fg);
   cmp_draw(r,c,nodeptr,"=");
   }

WORD equ_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 3;
   return(compile_node(nodeptr,*len,buf));
   }

void neq_init(nodeptr)
   struct logic_element *nodeptr;
   {
   cmp_node(nodeptr, "NEQ:  (1) != (2)", NEQ, NEQB, NEQ, NEQB, NEQ_OBJ);
   }

void neq_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void neq_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[NEQ_OBJ].label,fg,bg);
   coverlay(r,c,objs[NEQ_OBJ].overlay,fg);
   cmp_draw(r,c,nodeptr,"!=");
   }

WORD neq_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 3;
   return(compile_node(nodeptr,*len,buf));
   }

void grt_init(nodeptr)
   struct logic_element *nodeptr;
   {
   cmp_node(nodeptr, "GRT:  (1) > (2)", GRT, GRTB, LES, LESB, LES_OBJ);
   }

void grt_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void grt_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[GRT_OBJ].label,fg,bg);
   coverlay(r,c,objs[GRT_OBJ].overlay,fg);
   cmp_draw(r,c,nodeptr,">");
   }

WORD grt_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 3;
   return(compile_node(nodeptr,*len,buf));
   }

void geq_init(nodeptr)
   struct logic_element *nodeptr;
   {
   cmp_node(nodeptr, "GEQ:  (1) >= (2)", GEQ, GEQB, LEQ, LEQB, LEQ_OBJ);
   }

void geq_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void geq_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[GEQ_OBJ].label,fg,bg);
   coverlay(r,c,objs[GEQ_OBJ].overlay,fg);
   cmp_draw(r,c,nodeptr,">=");
   }

WORD geq_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 3;
   return(compile_node(nodeptr,*len,buf));
   }

void les_init(nodeptr)
   struct logic_element *nodeptr;
   {
   cmp_node(nodeptr, "LES:  (1) < (2)", LES, LESB, GRT, GRTB, GRT_OBJ);
   }

void les_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void les_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[LES_OBJ].label,fg,bg);
   coverlay(r,c,objs[LES_OBJ].overlay,fg);
   cmp_draw(r,c,nodeptr,"<");
   }

WORD les_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 3;
   return(compile_node(nodeptr,*len,buf));
   }

void leq_init(nodeptr)
   struct logic_element *nodeptr;
   {
   cmp_node(nodeptr, "LEQ:  (1) <= (2)", LEQ, LEQB, GEQ, GEQB, GEQ_OBJ);
   }

void leq_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void leq_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[LEQ_OBJ].label,fg,bg);
   coverlay(r,c,objs[LEQ_OBJ].overlay,fg);
   cmp_draw(r,c,nodeptr,"<=");
   }

WORD leq_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 3;
   return(compile_node(nodeptr,*len,buf));
   }

void mov_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str1[12], strdest[12];
   BYTE type1, typedest;
   DWORD reg1, regdest;

   get_values("MOV dest = Value",
              "Destination ? ",
              "Value       ? ",
              NULL,
              strdest,
              str1,
              NULL);
   if ((parse_io_str(str1,&type1,&reg1))
           && (parse_io_str(strdest,&typedest,&regdest)))
      {
      if ((typedest == VAR_TYPE)
             && ((type1 == VAR_TYPE) || (type1 == BYTEVAL_TYPE) || (type1 == WORDVAL_TYPE) || (type1 == DWORDVAL_TYPE)))
         {
         nodeptr->opcode[0] = (type1 == VAR_TYPE) ? MOV : MOVD;
         nodeptr->opcode[1] = regdest;
         nodeptr->opflags[1] = typedest;
         *(DWORD *)&nodeptr->opcode[2] = 0;  /* clear all bytes */
         if (type1 == VAR_TYPE)
            {
            nodeptr->opflags[2] = VAR_TYPE;
            nodeptr->opcode[2] = reg1 & 0xff;
            }
         else
            {
            nodeptr->opflags[2] = DWORDVAL_TYPE;  /* this always wants to be a DWORD */
            *(DWORD *)&nodeptr->opcode[2] = reg1;
            }
         }
      else
         beep();
      }
   }

void mov_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void mov_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[MOV_OBJ].label,fg,bg);
   coverlay(r,c,objs[MOV_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   smallstrout(r+TEXTHEIGHT,c,"=");
   showregtype(r+TEXTHEIGHT,c+2*TEXTWIDTH,nodeptr->opflags[2],*(DWORD *)&nodeptr->opcode[2]);
   }

WORD mov_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 3;  /* MOV is 3 bytes:  MOV <DEST> <R> */
   if (nodeptr->opcode[0] == MOV)
      return(compile_node(nodeptr,*len,buf));
   *len = 6;  /* MOVD is 6 bytes:  MOV <DEST> <DWORDVAL> */
   return(compile_endian_node(nodeptr,2,buf,4));
   }

void jmp_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("JMP label","Label number ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if (type == LABEL_TYPE)
         {
         nodeptr->opcode[0] = JMP;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void jmp_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void jmp_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   BYTE abuf[12];

   cstrout(r,c,objs[JMP_OBJ].label,fg,bg);
   coverlay(r,c,objs[JMP_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+2*TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD jmp_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   if (nodeptr->opflags[1] == UNUSED_TYPE)  /* input type not known */
      return(COMPILE_UNKNOWN_PARAMS);
   buf[0] = nodeptr->opcode[0];
   if (is_littleendian)
      {
      buf[1] = label_addresses[nodeptr->opcode[1]] & 0xff;  /* little-endian */
      buf[2] = label_addresses[nodeptr->opcode[1]] >> 8;
      }
   else
      {
      buf[1] = label_addresses[nodeptr->opcode[1]] >> 8;  /* big-endian */
      buf[2] = label_addresses[nodeptr->opcode[1]] & 0xff;
      }
   *len = 3;
   return(COMPILE_OK);
   }

void osr_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("ONESHOT","oneshot Register ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if (type == ONESHOT_TYPE)
         {
         nodeptr->opcode[0] = OSR;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void osr_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void osr_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[OSR_OBJ].label,fg,bg);
   coverlay(r,c,objs[OSR_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD osr_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 2;
   return(compile_node(nodeptr,*len,buf));
   }

void ctt_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE strr[12], strlimit[12];
   BYTE typer, typelimit;
   DWORD regr, reglimit;

   get_values("Counter FALSE->TRUE",
              "Counter register ? ",
              "Preset limit     ? ",
              NULL,
              strr,
              strlimit,
              NULL);
   if ((parse_io_str(strr,&typer,&regr))
           && (parse_io_str(strlimit,&typelimit,&reglimit)))
      {
      if (((typelimit == WORDVAL_TYPE) || (typelimit == BYTEVAL_TYPE))
                      && (typer == COUNTER_TYPE))
         {
         nodeptr->opcode[0] = CTT;
         nodeptr->opcode[1] = regr;
         nodeptr->opflags[1] = typer;
         nodeptr->opcode[2] = reglimit & 0xff;
         nodeptr->opflags[2] = WORDVAL_TYPE;  /* this always wants to be a WORD */
         nodeptr->opcode[3] = (reglimit >> 8);
         }
      else
         beep();
      }
   }

void ctt_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void ctt_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[CTT_OBJ].label,fg,bg);
   coverlay(r,c,objs[CTT_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   showregtype(r+TEXTHEIGHT,c,nodeptr->opflags[2],(DWORD)(*(WORD *)&nodeptr->opcode[2]));
   }

WORD ctt_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_endian_node(nodeptr,2,buf,2));
   }

void ctf_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE strr[12], strlimit[12];
   BYTE typer, typelimit;
   DWORD regr, reglimit;

   get_values("Counter TRUE->FALSE",
              "Counter register ? ",
              "Preset limit     ? ",
              NULL,
              strr,
              strlimit,
              NULL);
   if ((parse_io_str(strr,&typer,&regr))
           && (parse_io_str(strlimit,&typelimit,&reglimit)))
      {
      if (((typelimit == WORDVAL_TYPE) || (typelimit == BYTEVAL_TYPE))
                      && (typer == COUNTER_TYPE))
         {
         nodeptr->opcode[0] = CTF;
         nodeptr->opcode[1] = regr;
         nodeptr->opflags[1] = typer;
         nodeptr->opcode[2] = reglimit & 0xff;
         nodeptr->opflags[2] = WORDVAL_TYPE;  /* this always wants to be a WORD */
         nodeptr->opcode[3] = (reglimit >> 8);
         }
      else
         beep();
      }
   }

void ctf_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void ctf_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[CTF_OBJ].label,fg,bg);
   coverlay(r,c,objs[CTF_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   showregtype(r+TEXTHEIGHT,c,nodeptr->opflags[2],(DWORD)(*(WORD *)&nodeptr->opcode[2]));
   }

WORD ctf_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_endian_node(nodeptr,2,buf,2));
   }

void tmt_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE strr[12], strticks[12];
   BYTE typer, typeticks;
   DWORD regr, regticks;

   get_values("Timer TRUE",
              "Timer register ? ",
              "ticks (1ms)    ? ",
              NULL,
              strr,
              strticks,
              NULL);
   if ((parse_io_str(strr,&typer,&regr))
           && (parse_io_str(strticks,&typeticks,&regticks)))
      {
      if (((typeticks == WORDVAL_TYPE) || (typeticks == BYTEVAL_TYPE))
                      && (typer == TIMER_TYPE))
         {
         nodeptr->opcode[0] = TMT;
         nodeptr->opcode[1] = regr;
               /* EVEN dataregs are flags/accum, ODD are used for RTC storage */
         nodeptr->opflags[1] = typer;
         nodeptr->opcode[2] = regticks & 0xff;
         nodeptr->opflags[2] = WORDVAL_TYPE;  /* this always wants to be a WORD */
         nodeptr->opcode[3] = (regticks >> 8);
         }
      else
         beep();
      }
   }

void tmt_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void tmt_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[TMT_OBJ].label,fg,bg);
   coverlay(r,c,objs[TMT_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   showregtype(r+TEXTHEIGHT,c,nodeptr->opflags[2],(DWORD)(*(WORD *)&nodeptr->opcode[2]));
   }

WORD tmt_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_endian_node(nodeptr,2,buf,2));
   }

void tmf_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE strr[12], strticks[12];
   BYTE typer, typeticks;
   DWORD regr, regticks;

   get_values("Timer FALSE",
              "Timer register ? ",
              "ticks (1ms)    ? ",
              NULL,
              strr,
              strticks,
              NULL);
   if ((parse_io_str(strr,&typer,&regr))
           && (parse_io_str(strticks,&typeticks,&regticks)))
      {
      if (((typeticks == WORDVAL_TYPE) || (typeticks == BYTEVAL_TYPE))
                      && (typer == TIMER_TYPE))
         {
         nodeptr->opcode[0] = TMF;
         nodeptr->opcode[1] = regr;
               /* EVEN dataregs are flags/accum, ODD are used for RTC storage */
         nodeptr->opflags[1] = typer;
         nodeptr->opcode[2] = regticks & 0xff;
         nodeptr->opflags[2] = WORDVAL_TYPE;  /* this always wants to be a WORD */
         nodeptr->opcode[3] = (regticks >> 8);
         }
      else
         beep();
      }
   }

void tmf_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void tmf_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[TMF_OBJ].label,fg,bg);
   coverlay(r,c,objs[TMF_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   showregtype(r+TEXTHEIGHT,c,nodeptr->opflags[2],(DWORD)(*(WORD *)&nodeptr->opcode[2]));
   }

WORD tmf_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_endian_node(nodeptr,2,buf,2));
   }

void rtmt_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE strr[12], strticks[12];
   BYTE typer, typeticks;
   DWORD regr, regticks;

   get_values("Retentive Timer TRUE",
              "Timer register ? ",
              "ticks (1ms)    ? ",
              NULL,
              strr,
              strticks,
              NULL);
   if ((parse_io_str(strr,&typer,&regr))
           && (parse_io_str(strticks,&typeticks,&regticks)))
      {
      if (((typeticks == WORDVAL_TYPE) || (typeticks == BYTEVAL_TYPE))
                      && (typer == TIMER_TYPE))
         {
         nodeptr->opcode[0] = RTMT;
         nodeptr->opcode[1] = regr;
               /* EVEN dataregs are flags/accum, ODD are used for RTC storage */
         nodeptr->opflags[1] = typer;
         nodeptr->opcode[2] = regticks & 0xff;
         nodeptr->opflags[2] = WORDVAL_TYPE;  /* this always wants to be a WORD */
         nodeptr->opcode[3] = (regticks >> 8);
         }
      else
         beep();
      }
   }

void rtmt_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void rtmt_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[RTMT_OBJ].label,fg,bg);
   coverlay(r,c,objs[RTMT_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   showregtype(r+TEXTHEIGHT,c,nodeptr->opflags[2],(DWORD)(*(WORD *)&nodeptr->opcode[2]));
   }

WORD rtmt_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_endian_node(nodeptr,2,buf,2));
   }

void rtmf_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE strr[12], strticks[12];
   BYTE typer, typeticks;
   DWORD regr, regticks;

   get_values("Retentive Timer FALSE",
              "Timer register ? ",
              "ticks (1ms)    ? ",
              NULL,
              strr,
              strticks,
              NULL);
   if ((parse_io_str(strr,&typer,&regr))
           && (parse_io_str(strticks,&typeticks,&regticks)))
      {
      if (((typeticks == WORDVAL_TYPE) || (typeticks == BYTEVAL_TYPE))
                      && (typer == TIMER_TYPE))
         {
         nodeptr->opcode[0] = RTMF;
         nodeptr->opcode[1] = regr;
               /* EVEN dataregs are flags/accum, ODD are used for RTC storage */
         nodeptr->opflags[1] = typer;
         nodeptr->opcode[2] = regticks & 0xff;
         nodeptr->opflags[2] = WORDVAL_TYPE;  /* this always wants to be a WORD */
         nodeptr->opcode[3] = (regticks >> 8);
         }
      else
         beep();
      }
   }

void rtmf_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void rtmf_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[RTMF_OBJ].label,fg,bg);
   coverlay(r,c,objs[RTMF_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   showregtype(r+TEXTHEIGHT,c,nodeptr->opflags[2],(DWORD)(*(WORD *)&nodeptr->opcode[2]));
   }

WORD rtmf_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 4;
   return(compile_endian_node(nodeptr,2,buf,2));
   }

void reset_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;

   get_values("Reset timer/counter","timer/counter ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if ((type == TIMER_TYPE) || (type == COUNTER_TYPE))
         {
         nodeptr->opcode[0] = RESET;
         if (type == TIMER_TYPE)
            nodeptr->opcode[1] = reg;
               /* EVEN dataregs are flags/accum, ODD are used for RTC storage */
         else
            nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void reset_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void reset_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[RESET_OBJ].label,fg,bg);
   coverlay(r,c,objs[RESET_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD reset_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 2;
   return(compile_node(nodeptr,*len,buf));
   }

void inv_init(nodeptr)
   struct logic_element *nodeptr;
   {
   nodeptr->opcode[0] = INVERT;
   }

void inv_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void inv_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   cstrout(r,c,objs[INVERT_OBJ].label,fg,bg);
   coverlay(r,c,objs[INVERT_OBJ].overlay,fg);
   }

WORD inv_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 1;
   return(compile_node(nodeptr,*len,buf));
   }

void lbl_init(nodeptr)
   struct logic_element *nodeptr;
   {
   BYTE str[12];
   BYTE type;
   DWORD reg;
   WORD rung, index, found;

   for (found = FALSE, rung = 0; rung < MAX_RUNGS; rung++)
      {
      if (used_rungs[rung])
         {
         for (index = 0; index < rungs[rung].num_elements; index++)
            {
            if (nodeptr == &rungs[rung].node[index])
               {
               found = TRUE;
               break;   /* UGLY, UGLY KLUDGE! */
               }
            }
         if (found)
            break;  /* exit outer loop, too */
         }
      }
      /* I have to make sure this is the *FIRST* node on a *ROOT* rung, and
       * since the rung/index are not available, I have to find it.  UGLY!
       */
   if (index != 0)
      {
      nodeptr->objindex = NULL_OBJ;
      return(beep());
      }
   for (index = 0; index < ladder_size; index++)
      if (rung_list[index] == rung)
         break;
   if (index == ladder_size)  /* not a root rung */
      {
      nodeptr->objindex = NULL_OBJ;
      return(beep());
      }
   get_values("LABEL","Label number ? ",NULL,NULL,str,NULL,NULL);
   if (parse_io_str(str,&type,&reg))
      {
      if (type == LABEL_TYPE)
         {
         nodeptr->opcode[0] = LBL;
         nodeptr->opcode[1] = reg;
         nodeptr->opflags[1] = type;
         }
      else
         beep();
      }
   }

void lbl_del(r,c,nodeptr)
   WORD r,c;
   struct logic_element *nodeptr;
   {
   erase_node_notes(r,c);
   clear_node(nodeptr);
   }

void lbl_draw(r,c,nodeptr,fg,bg)
   WORD r,c;
   struct logic_element *nodeptr;
   WORD fg,bg;
   {
   BYTE abuf[12];

   cstrout(r,c,objs[LBL_OBJ].label,fg,bg);
   coverlay(r,c,objs[LBL_OBJ].overlay,fg);
   showregtype(r-TEXTHEIGHTSMALL,c+2*TEXTWIDTH,nodeptr->opflags[1],(DWORD)nodeptr->opcode[1]);
   }

WORD lbl_comp(nodeptr,len,buf)
   struct logic_element *nodeptr;
   WORD *len;
   BYTE *buf;
   {
   *len = 0;
   if (label_addresses[nodeptr->opcode[1]] != NULL_LABEL)
      {
      if (program_len)
         {
         if (label_addresses[nodeptr->opcode[1]] == program_len-1)
            return(COMPILE_OK);
         return(COMPILE_DUP_ADDR);
         }
      return(COMPILE_UNKNOWN_ADDR);
      }
   if (program_len == 0)
      return(COMPILE_UNKNOWN_ADDR);  /* close enough... */
   label_addresses[nodeptr->opcode[1]] = program_len - 1;
      /* the JMP target must be the SETLINETRUE at the start of this line */
   return(COMPILE_OK);
   }

void nullfn()  /* something to point to */
   {
   }
